/** This file is part of Reactor.
 *  Copyright (c) 2020-2024 Kedron Holdings LLC, All Rights Reserved.
 *  Reactor is not public domain or open source. Distribution or derivative works are expressly prohibited.
 */
/* globals feather */

export const version = 24119;

const required_core_version = 24115;  /* MUST sync to server/wsapi/wsapi.js */
const required_api_version = 24115;  /* Sync to client/ClientAPI.js */

const BUILD_VERSION = "userauth-24120-7745fb8d"; /* automatically set -- do not modify */

const LICENSE_VERSION = 23054;

import api from '/client/ClientAPI.js';

import * as util from './reactor-ui-common.js';
const idSelector = util.idSelector; /* freq use */

import alertWatcher from './alert-watcher.js';

import { _T, getLocale, _LD } from './i18n.js';

export default (function($) {

    var current_tab = false;

    function TBD() {
        util.showSysModal({ title: "Coming Soon!", body: "This function has not yet been implemented"});
    }

    function getAuthInfo() {
        let auth = {};
        const cookies = document.cookie.split( /;\s*/ ) || [];
        try {
            let cookie = cookies.find( el => el.match( /^reactor-session=/ ) );
            if ( cookie ) {
                let t = cookie.split( /=/ )[1] || "";
                t = t.split( /\|/ );
                if ( t.length >= 2 ) {
                    auth.user = t.shift();
                    auth.token = t.shift();
                    if ( t.length ) {
                        auth.applications = t.shift().split(',');
                        if ( 1 === auth.applications.length && "true" === auth.applications[0] ) {
                            // all applications permitted
                        } else if ( ! auth.applications.includes( 'reactor' ) ) {
                            window.location.href = "login.html?msg=Refused";
                            return auth;
                        }
                    }
                }
            }
        } catch ( err ) {
            console.error( err );
        }
        return auth;
    }

    function getMainNavItem( id, title, href, icon ) {
        var $el = $( '<li class="nav-item"> \
  <a class="nav-link tab-link" href="#"> \
    <span data-feather="cpu"></span> \
    <span class="nav-item-title">title</span> \
  </a> \
</li>' );
        $( 'span.nav-item-title', $el ).text( title );
        $( 'span[data-feather]', $el ).data('feather', icon).attr( 'data-feather', icon || 'cpu' );
        if ( "function" === typeof href ) {
            $( 'a.nav-link', $el ).on( 'click', href );
        } else {
            $( 'a.nav-link', $el ).attr( 'href', href || '#' );
        }
        $el.attr( 'id', 'nav-item-' + id );

        return $el;
    }

    function setActiveNav( id ) {
        $( '#nav-primary a.tab-link > span.sr-only' ).remove();
        $( '#nav-primary > li.nav-item a.tab-link' ).removeClass( 'active' );
        $( '#nav-primary > li#' + idSelector( 'nav-item-' + id ) + '.nav-item a.tab-link' )
            .addClass( 'active' );
        $( '#nav-primary a.tab-link.active' )
            .append( '<span class="sr-only"> (current)</span>' );
    }

    function switchTab( new_tab ) {
        var $el;
        if ( new_tab === current_tab ) {
            return false;
        }
        if ( current_tab ) {
            $el = $( 'div#' + idSelector( 'tab-' + current_tab ) + '.re-tab-container' );
            if ( false === $el.triggerHandler( 'suspend' ) ) {
                util.showSysModal( { title: _T('Operation in Progress'),
                    body: _T('Please complete the current operation before switching views'),
                } );
                console.log( `switchTab() tab ${current_tab} rejected request to switch` );
                return;
            }
            $el.hide();
        } else {
            /* We're lost, hide everything */
            $( 'div.re-tab-container' ).hide();
        }
        if ( new_tab ) {
            $el = $( 'div#' + idSelector( 'tab-' + new_tab ) + '.re-tab-container' );
            if ( 0 === $el.length ) {
                debugger;
                $el = $( 'div.re-tab-container' ).first();
            }
            $el.show();
            current_tab = new_tab;
            setActiveNav( new_tab );
            $el.triggerHandler( 'activate' );
        } else {
            current_tab = false;
        }

        /* Check core version */
        console.log( "Tab switch, checking core: have", api.core_version, "need", required_core_version );
        if ( api.isConnected() && ( api.core_version || 0 ) !== required_core_version ) {
            alert( "The cached UI in your browser needs to be updated. Please HARD REFRESH your browser or flush your browser cache. If this message persists, try restarting the Reactor core and then hard-refreshing." );
        }
    }

    function handleTabClick( event ) {
        var $el;
        var $target = $( event.target );
        var new_tab = ($target.closest( 'li.nav-item' ).attr( 'id' ) || 'status').replace( /nav-item-/, "" );
        switchTab( new_tab );
    }

    var doAboutTab = function( event ) {
        console.log("ReactorUICore.doAboutTab()");
        var $content = $( event.target );
        $content.empty();

        $( `<h2>${_T('#nav_about')}</h2>` ).appendTo( $content );
        $content.html('<h2>About Reactor</h2><p class="about-content">Reactor (Multi-hub) <span id="reactor-version"></span> &#169; 2020-2023 Kedron Holdings LLC, All Rights Reserved; Patrick Rigney (toggledbits), Principal Engineer.</p> \
<p style="font-size: 1.1rem; background-color: white; border: 2px dashed orange;" class="my-2 px-2 py-2 text-center">Reactor is currently a community-supported project, but it can only remain so if users make donations. If you haven\'t yet made a donation, please consider doing so.<br>\
<a href="https://buymeacoffee.com/toggledbits/" target="_blank"><img src="/img/buymeadram-209x48.png" width="209" height="48" alt="Buy me a dram!"></a> \
</p> \
<h4 id="license-header">License</h4>\
<div id="re-license"> \
<p>This license ("License") constitues the sole and entire agreement with respect to Reactor (Multi-hub) ("Reactor" or "Multi-System Reactor" or "MSR") between its author, Kedron Holdings, LLC ("KHLLC") as licensor and you ("You") as licensee. <b>YOUR POSSESSION AND/OR USE OF REACTOR CONSTITUTES AFFIRMATIVE AGREEMENT BY YOU TO ALL OF THE TERMS AND CONDITIONS OF THIS LICENSE AS FURTHER SPECIFIED BELOW.</b> If you do not agree to ALL of the terms and conditions herein, your sole remedy is to not use Reactor and destroy any and all copies you may have in your possession.</p> \
<p class="about-content"><b>RIGHTS AND RESTRICTIONS.</b> You are hereby granted a limited, non-exclusive, non-transferable license to use Reactor for non-commercial use while you remain in full compliance with all of terms and conditions herein. You may install and use Reactor only on compatible (as determined by KHLLC in its sole discretion) systems that you own and operate for your own private use. Except as otherwise expressly provided herein, You acquire no other rights or licenses. Reactor is and at all times remains the exclusive property of KHLLC. You shall not install, store, or use Reactor on a system owned by or for the benefit of any third-party, for any reason or purpose, without the prior express written permission of KHLLC. You may not use or distribute Reactor for any commercial purpose without the prior express written permission of KHLLC. You acknowledge and agree that Reactor is not "open source" software, even though versions of source code may be made available to You for its operation. You further agree that you shall not disseminate, disclose, disassemble, reverse engineer, or produce derivative works of Reactor, in whole or part. The Privacy Notice appearing later on this page is hereby incorporated herein and a part of this License.\
<p><b>DISCLAIMER OF WARRANTY.</b> <b>THIS SOFTWARE IS OFFERED AS-IS, TOGETHER WITH ANY AND ALL DEFECTS. KHLLC MAKES NO EXPRESS, IMPLIED, OR STATUTORY WARRANTY WITH RESPECT TO REACTOR INCLUDING (BUT NOT LIMITED TO) ANY WARRANTY WITH RESPECT TO MERCHANTABILITY, FITNESS, SECURITY, PRIVACY OR PERFORMANCE.</b> In the event of any failure of Reactor in any respect, your sole remedy is to notify KHLLC and/or discontinue use of Reactor. KHLLC may, in its sole discretion, fix any defect reported in any way it deems appropriate; however, nothing in this License creates any obligation or duty upon KHLLC to fix any defect, or update Reactor in any way, even if it is alleged or determined to be not conforming to its documentation, specifications, applicable standards, or best practices.\
<p><b>LIMITATION OF LIABILITY.</b> Under no circumstances will KHLLC or any of its agents, assigns, heirs, or successors be liable to you or any third party for any direct, indirect, special, indicental, or consequential damages whatsoever (including without limitation any loss of information or any other pecuniary loss) however caused and under any legal or equitable theory of liability, and whether or not for breach of contract, negligence or otherwise. You agree to indemnify, defend, and hold harmless KHLLC, its agents, assigns, heirs, and successors, from any of the foregoing arising in connection with your use of the Software. You bear solely the entire risk of the use of Reactor. Notwithstanding the foregoing, KHLLC\'s maximum liability under this agreement shall be the amount paid to KHLLC for Reactor. If the laws of your jurisdiction limit or prohibit any of the foregoing exclusions or limitations (including the preceding Disclaimer of Warranty), you may not use Reactor and your rights under this License terminate immediately without further notice.\
<p><b>GOVERNING LAW.</b> This License shall be governed by, and construed in accordance with, the laws of the State of Georgia, United States of America. Any legal action or proceeding arising instituted by You out of or relating to Reactor or this License shall be instituted in the courts of the State of Georgia and the parties hereto irrevocably submit to the jurisdiction of each such court in any action or proceeding. You hereby irrevocably waive and agree not to assert, by way of motion, as a defense, or otherwise, in every suit, action or other proceeding arising out of or based on Reactor or this License and brought in any such court, any claim that You are not subject personally to the jurisdiction of the above-named courts, that You are exempt or immune from attachment or execution, or that the venue of the suit, action, or proceeding is improper or inconvenient.\
<p><b>NOTICES.</b> Any notice to KHLLC with respect to Reactor or this License shall be made in writing delivered by United States Postal Service certified mail or commercial overnight carrier with signature confirmation of delivery to Kedron Holdings, LLC, Attn: Reactor License, 1029 North Peachtree Parkway #200, Peachtree City, GA 30269-6635. Any notice by KHLLC to You may be made by any means available to KHLLC (including but not limited to electronic mail or private messaging on forum web sites) and shall be deemed delivered when given.\
<p><b>MODIFICATION.</b> KHLLC reserves the right to modify the terms of this license at any time. Any such modifications may be made without prior announcement, and will be published within the software (as You see here). Your continued use of Reactor constitutes your acceptance of, and agreement to be bound by, the modifications. If you do not agree, your sole remedy is to discontinue use of Reactor. \
<p><b>TERMINATION.</b> This License terminates immediately upon any breach by You of any term or condition, without notice or prior warning required. This License also terminates when you cease any and all use of Reactor and have destroyed all copies in your possession (including backups), and may be terminated by KHLLC for any reason at any time without notice. Upon any termination of this License, all rights granted to You hereunder terminate immediately, and You must immediately destroy any and all copies, including backups, of Reactor in your possession. Except for the rights granted to You in "Rights and Restrictions" above, all other restrictions and provisions of this License shall survive any termination. \
<hr>\
<h5>Privacy Notice</h5>\
<p>The Reactor application and this web-based user interface collect information about the home automation gateways/hubs/controllers, devices and services that you configure, so the collection of this data is entirely under your control. The web interface uses cookies and local storage, but this data is not used elsewhere. This locally-stored data is thus also under your control to review or delete at any time you wish. The user interface loads some external resources/libraries from third-party sites, and you may be subject to data collection and tracking by these sites outside the control of KHLLC. Reactor gathers and reports general statistical, usage, and environment information to KHLLC, including information about the Reactor version in use, the hardware and software environment in which it is running, which hubs are used with it, configured language, region of the world, etc., but *does not* include personal identifying information such as email addresses, usernames, access tokens, passwords, etc. You explicitly grant permission for this data collection. You may at any time request a copy of this data, or that it be deleted, by written notice to KHLLC as provided in NOTICES above.</p> \
<p>On Reactor\'s bug tracking site and the KHLLC web site, IP addresses, cookies and local storage are used to track your session and preferences, and these sites load third-party libraries that may subject you to data collection and tracking by those third parties. You may request deletion of personal identifying information from KHLLC\'s web sites at any time. \
<p>Tracking information collected by and held with third parties may be reviewed or deleted by contacting the third party directly; KHLLC cannot and will not make these requests for you.</p> \
<p>We will use your email address only to communicate with you about the project, support, bugs, updates, and questions.  KHLLC does not and will not sell or intentionally disclose your email address or other personal identifying information to any third party without your express prior consent.</p> \
</div><!-- #re-license --> \
<div id="reactor-about-agreemnt" class="mb-3"></div> \
<h4>Host Public Key</h4> The key below is unique to your Reactor host system. Please provide it in its entirety (without editing and including the <tt>BEGIN</tt> and <tt>END</tt> lines) when requested for support. \
<pre id="host-public-key" class="user-select-all"></pre> \
<hr> \
<h4>Support and Updates</h4>\
<p>Discussions and announcements around Reactor (Multi-hub) are conducted in the Smart Home Community at <a href="https://smarthome.community/topic/768/category-topic-guide-read-before-posting" target="_blank">https://smarthome.community/</a>. Please <strong>do not</strong> use the Vera/eZLO forums for Reactor (Multi-Hub) discussions (they may ban you). If you have not yet signed up for an account at the Smart Home Community, please do. Please <em><a href="https://smarthome.community/topic/768/category-topic-guide-read-before-posting" target="_blank">read the posting guidelines</a></em> before asking for help from the community. Bugs may be reported in the Smart Home Community or using the <a href="https://reactor.toggledbits.com/mantisbt/" target="_blank">Mantis Bug Tracker</a> set up for the project (contact <a href="https://smarthome.community/user/toggledbits" target="_blank">@toggledbits</a> for account setup). Please <strong>do not</strong> email or DM the author directly for support &mdash; the author will only respond to community public posts and Mantis bug reports. Recognized annual/monthly donors receive priority support.<sup>1</sup></p> \
<hr> \
<h4>Open Source Licenses</h4> \
<p>Reactor uses and may in its distribution include all or part of the following open source products/libraries:</p> \
<ul id="opensourceproducts"> \
<li><a href="https://github.com/nodejs/node" target="_blank">nodejs</a> - Copyright 2012-2020 by various contributors; <a href="https://github.com/nodejs/node/blob/master/LICENSE" target="_blank">full notice and (MIT) license</a>.</li> \
<li><a href="https://github.com/influxdata/influxdb-client-js" target="_blank">@influxdata/influxdb-client</a> - Copyright (c) 2019 InfluxData; <a href="https://github.com/influxdata/influxdb-client-js/blob/master/LICENSE" target="_blank">(MIT) license</a></li> \
<li><a href="https://www.npmjs.com/package/@mdi/font" target="_blank">@mdi/font</a> (Material Design icons) - <a href="https://github.com/Templarian/MaterialDesign-Webfont/blob/master/LICENSE" target="_blank">Pictogrammers Free License</a></li> \
<li><a href="https://www.npmjs.com/package/@popperjs/core" target="_blank">@popperjs/core</a> - Copyright (c) 2021 Floating UI contributors; <a href="https://github.com/floating-ui/floating-ui/blob/master/LICENSE" target="_blank">MIT License</a></li> \
<li><a href="https://github.com/expressjs/body-parser" target="_blank">body-parser</a> - Copyright (c) 2014 Jonathan Ong <me@jongleberry.com>; Copyright (c) 2014-2015 Douglas Christopher Wilson <doug@somethingdoug.com>; <a href="https://github.com/expressjs/body-parser/blob/master/LICENSE" target="_blank">(MIT) license</a></li> \
<li><a href="https://github.com/twbs/icons" target="_blank">Bootstrap Icons</a> - Copyright (c) 2019-2020 The Bootstrap Authors; <a href="https://github.com/twbs/icons/blob/main/LICENSE.md" target="_blank">full notice and (MIT) license</a></li> \
<li><a href="https://getbootstrap.com/" target="_blank">Bootstrap</a> - Copyright (c) 2011-2021 Twitter, Inc.; Copyright (c) 2011-2021 The Bootstrap Authors; <a href="https://tldrlegal.com/license/mit-license" target="_blank">MIT license</a>.</li> \
<li><a href="https://www.npmjs.com/package/digest" target="_blank">digest</a> - Copyright (c) 2011 Fractal <contact@wearefractal.com>; <a href="https://github.com/wearefractal/node-digest/blob/master/LICENSE" target="_blank">MIT license</a>.</li> \
<li><a href="https://github.com/iximiuz/node-diskusage-ng" target="_blank">diskusage-ng</a> - Copyright (c) 2015 Ivan Velichko <iximiuz@gmail.com>; <a href="https://github.com/iximiuz/node-diskusage-ng/blob/master/LICENSE.txt" target="_blank">(MIT) license</a></li> \
<li><a href="https://expressjs.com/" target="_blank">express</a> - Copyright (c) 2009-2014 TJ Holowaychuk <tj@vision-media.ca>; Copyright (c) 2013-2014 Roman Shtylman <shtylman+expressjs@gmail.com>; Copyright (c) 2014-2015 Douglas Christopher Wilson <doug@somethingdoug.com>; <a href="https://github.com/expressjs/express/blob/master/LICENSE" target="_blank">(MIT) license</a></li> \
<li><a href="https://feathericons.com/" target="_blank">feather-icons</a> - Copyright (c) 2013-2017 Cole Bemis; <a href="https://github.com/feathericons/feather/blob/master/LICENSE" target="_blank">(MIT) license</a></li> \
<li><a href="http://fontawesome.io/" target="_blank">font-awesome</a> - <a href="https://github.com/FortAwesome/Font-Awesome/blob/master/LICENSE.txt" target="_blank">Font Awesome Free License</a></li> \
<li><a href="https://github.com/gridstack/gridstack.js#readme" target="_blank">gridstack.js</a> - Copyright (c) 2021 Alain Dumesny; <a href="https://github.com/gridstack/gridstack.js/blob/master/LICENSE" target="_blank">(MIT) license</a></li> \
<li><a href="https://jquery-ui.com" target="_blank">jQuery-UI</a> - Copyright OpenJS Foundation and jQuery contributors; <a href="https://tldrlegal.com/license/mit-license" target="_blank">MIT license</a>.</li> \
<li><a href="https://jquery.com" target="_blank">jQuery</a> - Copyright OpenJS Foundation and jQuery contributors; <a href="https://tldrlegal.com/license/mit-license" target="_blank">MIT license</a>.</li> \
<li><a href="https://github.com/nodeca/js-yaml#readme" target="_blank">js-yaml</a> - Copyright (C) 2011-2015 by Vitaly Puzrin; <a href="https://github.com/nodeca/js-yaml/blob/master/LICENSE" target="_blank">(MIT) license</a></li> \
<li><a href="https://github.com/node-fetch/node-fetch" target="_blank">node-fetch</a> - Copyright (c) 2016 - 2020 Node Fetch Team; <a href="https://github.com/node-fetch/node-fetch/blob/master/LICENSE.md" target="_blank">(MIT) license</a></li> \
<li><a href="https://www.npmjs.com/package/node-stream-zip" target="_blank">node-stream-zip</a> - Copyright (c) 2021 <a href="https://github.com/antelle" target="_blank">Antelle</a>; <a href="https://github.com/antelle/node-stream-zip/blob/master/LICENSE" target="_blank">license</a></li> \
<li><a href="https://nodemailer.com/" target="_blank">nodemailer</a> - Copyright (c) 2011-2019 Andris Reinman; <a href="https://github.com/nodemailer/nodemailer/blob/master/LICENSE" target="_blank">license</a></li> \
<li><a href="https://github.com/expressjs/serve-static#readme" target="_blank">serve-static</a> - Copyright (c) 2010 Sencha Inc.; Copyright (c) 2011 LearnBoost; Copyright (c) 2011 TJ Holowaychuk; Copyright (c) 2014-2016 Douglas Christopher Wilson; <a href="https://github.com/expressjs/serve-static/blob/master/LICENSE" target="_blank">(MIT) license</a></li> \
<li><a href="https://github.com/uuidjs/uuid#readme" target="_blank">uuid</a> - Copyright (c) 2010-2020 Robert Kieffer and other contributors; <a href="https://github.com/uuidjs/uuid/blob/master/LICENSE.md" target="_blank">(MIT) license</a></li> \
<li><a href="https://github.com/websockets/ws" target="_blank">ws</a> - Copyright (c) 2011 Einar Otto Stangvik <einaros@gmail.com>; <a href="https://github.com/websockets/ws/blob/master/LICENSE" target="_blank">(MIT) license</a></li> \
</ul> \
<p><small>Persons and/or business entities are named in the foregoing list for reference and compliance only and are not endorsements by or of Reactor or KHLLC.</small></p> \
<p><small><sup>1</sup> All support is offered on an as-available basis. There is no guaranteed response time for support, but I will prioritize requests from users who have donated US$25 or more in the last 12 months.</small></p> \
');

        let cookies = document.cookie.split( /;\s*/ ) || [];
        let cookie = cookies.find( el => el.match( /^reactor-about=/ ) );
        let $ag = $( '#reactor-about-agreemnt', $content );
        if ( cookie && cookie.split( /=/ )[1] == LICENSE_VERSION ) {
            console.log( "cookie", cookie.split( /=/ )[1], "expecting", LICENSE_VERSION );
            $ag.html( '<div class="re-license-agreed">Thank you for using Reactor! To review the software license, please click the right-pointing arrow above.</div>' );
            let $lel = $( 'div#re-license' );
            $lel.hide();
            let $leh = $( 'h4#license-header' );
            $leh.append( '<span>&nbsp;<i class="bi bi-chevron-right ms-1"></i></span>' );
            $( 'span', $leh ).on( 'click', ev => {
                if ( $( 'span > i.bi', $leh ).hasClass( 'bi-chevron-right' ) ) {
                    $( 'span > i.bi', $leh ).removeClass( 'bi-chevron-right' ).addClass( 'bi-chevron-down' );
                    $lel.show();
                    $( 'div.re-license-agreed' ).text( 'Thank you for using Reactor!' );
                } else {
                    $( 'span > i.bi', $leh ).removeClass( 'bi-chevron-down' ).addClass( 'bi-chevron-right' );
                    $lel.hide();
                }
            });
        } else {
            $( '<div class="re-license-pending">To stop this About page from displaying first, please enter the last four characters of your <em>Host Public Key</em> (shown below). \
  <div class="row gy-2 gx-3 align-items-center"> \
    <div class="col-auto"> \
      <label class="form-label" for="hpk4">Last four characters of Host Public Key:</label> \
    </div> \
    <div class="col-auto"> \
      <input id="hpk4" class="form-control text-center ms-2" value="" size="6"> \
    </div> \
    <div class="col-auto"> \
      <button id="re-agree" class="btn btn-primary ms-2">Verify</button> \
    </div> \
  </div> \
</div>' )
                .appendTo( $ag );
            $( 'button#re-agree', $content ).on( 'click', async () => {
                let hk = await api.getHostPublicKey();
                if ( "string" === typeof hk ) {
                    hk = hk.replace( /[-]+END PUBLIC KEY.*/, "" ).replace( /[\r\n]/g, "" );
                    if ( $( 'input#hpk4', $content ).val().trim() === hk.substring( hk.length - 4 ) ) {
                        const d = new Date();
                        d.setTime( d.getTime() + 90 * 86400000 );  /* 90-day cookie life */
                        document.cookie = `reactor-about=${LICENSE_VERSION}; expires=${d.toUTCString()}; path=/reactor`;
                        switchTab( 'status' );
                    } else {
                        document.cookie = `reactor-about=${LICENSE_VERSION}; expires=${(new Date(0)).toUTCString()}; path=/reactor`;
                        util.showSysModal({title:'Invalid Entry',body:'Your entry did not match the last four characters of the Host Public Key. Please try again.'});
                    }
                } else {
                    document.cookie = `reactor-about=${LICENSE_VERSION}; expires=${(new Date(0)).toUTCString()}; path=/reactor`;
                    util.showSysModal({title:'Sorry!',body:'I was not able to retrieve the Host Public Key. Please make sure the Reactor service is up and running.'});
                }
            });
        }

        $( 'span#reactor-version', $content ).text( String(BUILD_VERSION) );
        api.getHostPublicKey().then( key => {
            $( '#host-public-key', $content ).text( key );
        }).catch( err => {
            $( '#host-public-key', $content ).html( '<em>Not available! Please make sure the Reactor service is running!</em>' );
            console.error( err );
        });

        /* Attribution/version info for controllers */
        const controllers = api.getControllers();
        let $ct = $( 'div#about-controllers ul' );
        for ( let id of controllers ) {
            let ctrl = api.getController( id );
            if ( ctrl.metadata?.pkginfo ) {
                if ( 0 === $ct.length ) {
                    $ct = $( '<div id="about-controllers"></div>' )
                        .insertAfter( $( '#host-public-key' ) );
                    $( '<hr/>' ).insertBefore( $ct );
                    $ct.append( '<h4>Controllers and Plugins</h4>' );
                    $ct = $ct.append( '<ul></ul>' );
                }
                let $li = $( '<li></li>' ).appendTo( $( 'ul', $ct ) );
                let m = ctrl.metadata;
                $li.text( `${m.implementationName} [${m.pkginfo.version}] ${m.pkginfo.author}` );
                if ( m.pkginfo.license ) {
                    $li.append( `<span>; ${m.pkginfo.license}</span>` );
                }
                if ( m.pkginfo.homepage ) {
                    let $cp = $( '<span>; </span>' ).appendTo( $li );
                    $( '<a></a>' )
                        .attr( 'href', m.pkginfo.homepage )
                        .attr( 'target', '_blank' )
                        .text( m.pkginfo.homepage )
                        .appendTo( $cp );
                }
            }
        }

        /* Attribution for language file */
        if ( "string" === typeof _LD.attribution ) {
            let $l = $( '<li></li>' ).prependTo( 'ul#opensourceproducts', $content );
            if ( _LD.github_repository ) {
                $l = $( '<a></a>').attr( {
                    href: 'https://github.com/' + _LD.github_repository,
                    target: '_blank'
                }).appendTo( $l );
            }
            $l.text( _LD.attribution );
        }
    };

    var doToolsTab = function( event ) {
        var $content = $( event.target );
        $content.empty();

        $( `<h2>${_T('#nav_tools')}</h2>`).appendTo( $content );
        let $ct = $( '<div></div>' ).appendTo( $content );
        $( `<button class="btn btn-sm btn-danger"><i class="bi bi-bootstrap-reboot pe-1"></i>${_T('Restart Reactor')}</button>` )
            .on( 'click', ( ev ) => {
                const $el = $( ev.currentTarget );
                $el.prop( 'disabled', true );
                setTimeout( () => $el.prop( 'disabled', false ), 5000 );
                const auth = getAuthInfo();
                $.ajax({
                    url: "/api/v1/reload",
                    timeout: 10000,
                    beforeSend: function( xhr ) { if (auth?.token) xhr.setRequestHeader("Authorization","Bearer "+auth.token); }
                }).done( () => {
                    console.log("Reload server request completed.");
                }).fail( a => {
                    console.error( a );
                    if ( 401 === a.status || 403 === a.status ) {
                        window.location.href = 'login.html';
                    }
                });
            })
            .appendTo( $ct );

/*
        $ct = $( '<div></div>' ).appendTo( $content );
        $( '<button class="btn btn-sm btn-primary">Test</button>' )
            .on( 'click', async function() {
                try {
                    let mod = await import( './entity-picker.js' );
                    let picker = mod.default;
                    picker.pick().then( (thing) => { alert("picked "+thing); } );
                } catch ( err ) {
                    alert( err );
                    console.log( err );
                }
            })
            .appendTo( $ct );
*/

        $ct = $( '<footer></footer>' ).appendTo( $content );
        $( '<span></span>' )
            .text( `${_LD.locale} (${_LD.description}) ${_LD.version}; browser ${navigator.userAgent} lang ${navigator.languages.join(', ')}` )
            .appendTo( $ct );
    };


    var start = function() {
        /* Special handlers to assist tabs, since <div> elements do not
           normally get show/hide events in jQuery */
        $.each(['show', 'hide'], function (i, ev) {
            var el = $.fn[ev];
            $.fn[ev] = function () {
            this.trigger(ev);
            return el.apply(this, arguments);
            };
        });

        var $main = $( 'main[role="main"]' ).first().empty();
        $( '<div class="re-ui-alerts collapse"></div>' ).appendTo( $main );
        $main = $( '<div class="re-ui-content"></div>' ).appendTo( $main );
        var $ct = $( 'ul#nav-primary' ).empty();

        /* Set up sub-modules and navigation */
        getMainNavItem( 'status', _T('#nav_status'), handleTabClick, 'activity' )
            .appendTo( $ct );
        const readyStatusTab = new Promise( ( resolve ) => {
            import( './reactor-ui-status.js' ).then( function( mod ) {
                mod.default.init( $main );
                resolve();
            }).catch( function( err ) {
                console.log(err);
                console.trace(err);
            });
        });

        /* Rules has subnav, so a bit more complex construction */
        var $el = $( `<li class="nav-item"> \
  <div class="nav-group expandable closed d-flex flex-row align-items-center"> \
    <div class="me-auto">\
      <a href="#" class="nav-link tab-link">\
        <span data-feather="cpu"></span> \
        <span class="nav-item-title">title</span></a> \
      </a>\
    </div>\
    <div class="nav-item-controls me-2">\
      <a id="nav-addruleset" class="d-flex align-items-center text-muted" href="#" aria-label="${_T('Add a new rule set')}">\
        <span data-feather="plus-circle"></span>\
      </a>\
    </div>\
  </div> \
</li>` );
        $( 'span.nav-item-title', $el ).text( _T( "#nav_rulesets" ) );
        $el.attr( 'id', 'nav-item-rules' );
        $( '<ul id="ruleset-nav" class="nav hide"></ul>' ).appendTo( $el ); /* subnav */
        $el.appendTo( $ct );
        import( './reactor-ui-rules.js' ).then( function( mod ) {
            mod.default.init( $main );
            $( 'li#nav-item-rules a.nav-link' ).on( 'click', handleTabClick );
        }).catch( function( err ) {
            console.error(err);
            console.trace(err);
        });

        getMainNavItem( 'reactions', _T('#nav_reactions'), handleTabClick, 'zap' )
            .appendTo( $ct );
        import( './reactor-ui-reactions.js' ).then( function( mod ) {
            mod.default.init( $main );
        }).catch( function( err ) {
            console.log(err);
        });

        getMainNavItem( 'expressions', _T('#nav_expressions'), handleTabClick, 'divide' )
            .appendTo( $ct );
        import( './reactor-ui-expressions.js' ).then( function( mod ) {
            mod.default.init( $main );
        }).catch( function( err ) {
            console.log(err);
        });

        getMainNavItem( 'scope', _T('#nav_scope'), handleTabClick, 'eye' )
            .appendTo( $ct );
        import( './reactor-ui-scope.js' ).then( function( mod ) {
            mod.default.init( $main );
        }).catch( function( err ) {
            console.log(err);
        });

        getMainNavItem( 'entities', _T('#nav_entities'), handleTabClick, 'list' )
            .appendTo( $ct );
        import( './reactor-ui-entities.js' ).then( function( mod ) {
            mod.default.init( $main );
        }).catch( function( err ) {
            console.log(err);
            console.trace(err);
        });

        $( '<div id="tab-tools" class="re-tab-container">TBD</div>' )
            .hide()
            .on( 'show', doToolsTab )
            .appendTo( $main );
        getMainNavItem( 'tools', _T('#nav_tools'), handleTabClick, 'tool' )
            .appendTo( $ct );

        getMainNavItem( 'dashboard', _T('#nav_dashboard'), '/dashboard/', 'tablet' )
            .appendTo( $ct )
            .find( 'a.nav-link' ).attr( 'target', '_blank' );

        getMainNavItem( 'docs', _T('#nav_docs'), '/docs/', 'book-open' )
            .appendTo( $ct )
            .find( 'a.nav-link' ).attr( 'target', '_blank' );

        getMainNavItem( 'bugs', _T('#nav_bugs'),
                'https://reactor.toggledbits.com/mantisbt/', 'cloud-rain' )
            .appendTo( $ct )
            .find( 'a.nav-link' ).attr( 'target', '_blank' );

        getMainNavItem( 'donate', _T('#nav_donate'), 'https://buymeacoffee.com/toggledbits',
                'star' )
            .appendTo( $ct )
            .find( 'a.nav-link' ).attr( 'target', '_blank' );

        /* About */
        $( '<div id="tab-about" class="re-tab-container"></div>' )
            .hide()
            .on( 'activate', doAboutTab )
            .appendTo( $main );
        getMainNavItem( 'about', _T('#nav_about'), handleTabClick, 'info' )
            .appendTo( $ct );

        const cookies = document.cookie.split( /;\s*/ ) || [];
        try {
            const auth = getAuthInfo();
            if ( auth?.token ) {
                getMainNavItem( 'logout', _T(['#nav_logout','Log Out']), 'login.html', 'log-out' )
                    .appendTo( $ct );
            }
        } catch ( err ) {
            console.error( err );
        }

        /* Feather up */
        try {
            feather.replace();
        } catch( e ) {
            console.log("feather:");
            console.log(e);
        }

        $( 'a#re-api-connstate' ).text("Starting...");
        api.on( 'connected.core', function() {
            $( 'a#re-api-connstate' ).html( `<span class="text-success">${_T('Connected')}</span>` );
        });
        api.on( 'disconnected.core', function() {
            $( 'a#re-api-connstate' ).html( `<span class="text-danger"><b>${_T('DISCONNECTED!')}</b></span>` );
        });

        /* Version string in greeting/profile spot for now */
        $( '#site-greeting' ).text( BUILD_VERSION ).on( 'click.reactor', () => {
            switchTab( 'about' );
        });

        try {
            const d = new Date();
            if ( ( d.getMonth() === 11 && d.getDate() >= 8 ) || ( d.getMonth() === 0 && d.getDate() < 3 ) ) {
                $( 'div#re-nav-banner' )
                    .addClass( 'text-center' )
                    .css( 'color', '#0c0 !important' )
                    .text( "Happy Holidays and Best Wishes for a Peaceful and Prosperous New Year! -- @toggledbits" );
            }
        } catch ( err ) {
            console.error( err );
        }

        $( 'a#re-ui-alarms' ).on( 'click.reactor', () => {
            switchTab( 'status' );
        });

        api.off( 'auth_fail.core' ).on( 'auth_fail.core', () => {
            window.location.href = 'login.html';
        });
        api.start().then( function() { // no options passed; reusing prior authorization from reactor-ui.js
            /* API started before us, so before on handler applied */
            $( 'a#re-api-connstate' ).html( `<span class="text-success">${_T('Connected')}</span>` );

            /* Link up the alerts */
            alertWatcher.start();

            /* Launch the status page to start once the API is ready. */
            $( 'main > h1' ).remove();
            let cookie = cookies.find( el => el.match( /^reactor-about=/ ) );
            if ( cookie && cookie.split( /=/ )[1] == LICENSE_VERSION ) {
                readyStatusTab.then( () => {
                    switchTab( 'status' );
                });
            } else {
                switchTab( 'about' );
            }

            /* Display/update host time */
            /* Note: browser time is updated by reactor-ui.js */
            (function (){
                let host_last = 0;
                let host_offset = null;
                let host_tz = "unknown";
                let $el = $( 'a#re-hostclock' );
                let request_pending = false;
                $el.on( 'click.reactor', () => { host_last = 0; } );
                let update_host_time = function() {
                    let now = new Date();
                    let d = now - host_last;
                    if ( d >= 900000 && api.isConnected() && !request_pending ) {
                        // Figure out the current RTT time for a query
                        request_pending = true;
                        const auth = getAuthInfo();
                        $.ajax( `/api/v1/systime?rtt=${now.getTime()}`, {
                            dataType: 'json',
                            beforeSend: function( xhr ) { if (auth?.token) xhr.setRequestHeader("Authorization","Bearer "+auth.token); }
                        }).done( data => {
                            now = new Date(); // freshen
                            host_last = now.getTime();
                            /* Compute offset with skew for timezone difference between host (engine) and browser */
                            host_offset = data.engine_time - now.getTime();
                            host_offset += 60000 * ( now.getTimezoneOffset() + data.tzoffset );
                            host_tz = data.timezone;
                            console.log("systime: host", data.engine_time, ", me", host_last, "; host_offset=",
                                host_offset, "; full response", data);
                            request_pending = false;
                        }).fail( (e) => {
                            console.error("Host time query", e);
                            host_last = 0;
                            host_offset = null;
                            request_pending = false;
                        });
                    }

                    /* Fixed format time display */
                    if ( null !== host_offset && api.isConnected() ) {
                        const host_time = now.getTime() + host_offset;
                        $el.text( new Date( host_time ).toLocaleTimeString( getLocale(), { hour12: false } ) );
                        $( 'a#re-hostclock' ).closest( 'span' ).attr( 'title', _T( 'Host time' ) + ` (${host_tz})` );
                        /* Update timing based on host offset */
                        setTimeout( update_host_time, 1000 - (host_time % 1000 ) );
                    } else {
                        $el.text( "--:--:--" );
                        setTimeout( update_host_time, 1000 );
                    }
                };
                update_host_time();
            })();
            $( 'a#re-hostclock' ).closest( 'span' ).attr( 'title', _T( 'Host time' ) );
            $( 'a#re-sysclock' ).closest( 'span' ).attr( 'title', _T( 'Local (browser) time' ) +
                ` (${String(Intl?.DateTimeFormat()?.resolvedOptions()?.timeZone)})` );
        }).catch( err => {
            if ( err.status === 401 ) {
                window.location.href = "login.html";
            } else {
                console.error( err );
            }
        });
    };

    return {
        "start": start
    };
})(jQuery);
