/** This file is part of Reactor.
 *  Copyright (c) 2020-2024 Kedron Holdings LLC, All Rights Reserved.
 *  Reactor is not public domain or open source. Distribution or derivative works are expressly prohibited.
 */
/* globals bootstrap */

export const version = 24101;

import api from '/client/ClientAPI.js';

import * as util from './reactor-ui-common.js';

/* Frequently used */
var isEmpty = util.isEmpty;
var idSelector = util.idSelector;

import { _T } from './i18n.js';

/* ??? needs conversion to Tab class */

export default (function($) {

    var tabId = "tab-entities";

    function clean( str ) {
        try {
            if ( "boolean" === typeof str ) {
                str = _T( String( str ) );
            } else if ( null === str || "undefined" === typeof str ) {
                str = _T( 'null' );
            } else {
                str = JSON.stringify( str );
            }
        } catch( err ) {
            // console.log("Can't stringify",str,":",err);
            str = String( str );
        }
        return str.replace( /\n/g, "\\n" ).replace( /\r/g, "\\r" ).replace( /\t/g, "\\t" ).replace( /[\x00-\x1f\x7f]/g,
            function( m ) {
                return '\\x' + m.charCodeAt( 0 ).toString( 16 );
            } );
    }

    /** Copied from reaction-editor.js; need to just use it where it is. Although this is MODIFIED */
    function getArgumentField( action, pname, parm, $container, lastval ) {
        var $inp, $opt;
        parm.type = parm.type || "string";
        if ( "undefined" !== typeof parm.value ) {
            $inp = $( '<input type="hidden" class="argument">' ).val( String( parm.value ) );
        } else if ( null !== ( parm.values || null ) ) {
            let values = parm.values;
            if ( "string" === typeof values ) {
                /* String to array of values */
                values = values.split( /, */ );
            } else if ( ! Array.isArray( values ) ) {
                /* Convenience format of capability: params as dict rather than array of key/value pairs */
                if ( "object" === typeof values ) {
                    let m = [];
                    Object.keys( values ).forEach( key => { m.push( { [key]: values[key] } ); } );
                    values = m;
                } else {
                    /* Ugly, but not crashy */
                    values = [ String( values ) ];
                }
            }
            if ( undefined !== window.HTMLDataListElement ) {
                /* Use datalist when supported (allows more flexible entry) */
                var dlid = ("dl-" + action + '-' + pname).replace( /[^a-z0-9-]/ig, "-" );
                if ( 0 === $( 'datalist#' + idSelector( dlid ) ).length ) {
                    /* Datalist doesn't exist yet, create it */
                    $inp = $('<datalist class="argdata"></datalist>').attr( 'id', dlid );
                    values.forEach( val => {
                        $opt = $( '<option></option>' );
                        if ( null !== val && "object" === typeof( val ) ) {
                            Object.keys( val ).forEach( z => {
                                $opt.val( String( z ) );
                                $opt.text( String( val[z] ) + ( parm.default === val[z] ? " *" : "" ) );
                            });
                        } else {
                            $opt.val( String( val ) );
                            $opt.text( String( val ) + ( val === parm.default ? " *" : "" ) );
                        }
                        $inp.append( $opt );
                    });
                    /* Add variables and append to tab (datalists are global to tab) */
                    if ( false && ! parm.novars ) { /* MODIFIED */
                        this.appendVariables( $inp );
                    }
                    ( $container || this.$parent ).append( $inp );
                }
                /* Now pass on the input field */
                $inp = $( '<input class="argument form-control form-control-sm">' )
                    .attr( 'placeholder', _T('Click for predefined values' ) )
                    .attr( 'list', dlid );
                if ( undefined !== lastval ) {
                    $inp.val( lastval );
                } else if ( undefined !== parm.default ) {
                    $inp.val( String( parm.default ) );
                }
            } else {
                /* Standard select menu */
                $inp = $( '<select class="argument form-select form-select-sm"></select>' );
                if ( parm.optional ) {
                    $inp.append( `<option value="">${_T('(not specified)')}</option>` );
                }
                values.forEach( val => {
                    $opt = $( '<option></option>' );
                    if ( null !== val && "object" === typeof( val ) ) {
                        Object.keys( val ).forEach( z => {
                            $opt.val( String( z ) );
                            $opt.text( String( val[z] + ( parm.default === val ? " *" : "" ) ) );
                        });
                    } else {
                        $opt.val( String( val ) );
                        $opt.text( String( val ) + ( parm.default === val ? " *" : "" ) );
                    }
                    $inp.append( $opt );
                });
                /* Add variables */
                if ( false && ! parm.novars ) { /* MODIFIED */
                    this.appendVariables( $inp );
                }
                /* As a default, just choose the first option, unless specified & required */
                if ( undefined !== lastval ) {
                    $opt = $( `option[value="${lastval}"]`, $inp );
                    if ( 0 === $opt.length ) {
                        $( '<option></option>' ).val( lastval ).text( lastval ).prependTo( $inp );
                    }
                    $inp.val( lastval );
                } else if ( undefined !== parm.default && !parm.optional ) {
                    $inp.val( parm.default );
                } else {
                    $( 'option:first', $inp ).prop( 'selected', true );
                }
            }
            /* END MODIFIED */
        } else if ( parm.type.match( /^bool(ean)?$/ ) ) { /* MODIFIED */
            /* Menu */
            $inp = $('<select class="argument form-select form-select-sm"></select>');
            if ( parm.optional ) {
                $inp.prepend( `<option value="">${_T('(not specified)')}</option>` );
            }
            $inp.append( `<option value="false">${_T('false')}${"false"===String(parm.default)?" *":""}</option>` );
            $inp.append( `<option value="true">${_T('true')}${"true"===String(parm.default)?" *":""}</option>` );
            /* Add variables */
            if ( false && !parm.novars ) { /* MODIFIED */
                this.appendVariables( $inp );
            }
            /* Force default when available and not optional, otherwise first */
            if ( undefined !== lastval ) {
                $inp.val( lastval );
            } else if ( undefined !== parm.default && ! parm.optional ) {
                $inp.val( String( parm.default ) );
            } else {
                $( 'option:first', $inp ).prop( 'selected', true );
            }
        } else if ( false && parm.type === "ui1" && parm.min !== undefined && parm.max !== undefined ) {
            $inp = $('<div class="argument tbslider"></div>');
            $inp.slider({
                min: parm.min, max: parm.max, step: parm.step || 1,
                range: "min",
                stop: function ( ev, ui ) {
                    // DeusExMachinaII.changeDimmerSlider( $(this), ui.value );
                },
                slide: function( ev, ui ) {
                    $( 'a.ui-slider-handle', $( this ) ).text( ui.value );
                },
                change: function( ev, ui ) {
                    $( 'a.ui-slider-handle', $( this ) ).text( ui.value );
                }
            });
            $inp.slider("option", "disabled", false);
            $inp.slider("option", "value", parm.default || parm.min || 0 ); /* MODIFIED */
        } else if ( parm.type.match( "^(real|uint|int)$" ) || parm.type.match(/^(u?i)[1248]$/i ) ) {
            $inp = $( '<input class="argument narrow form-control form-control-sm">' );
            if ( ! parm.novars ) {
                $inp.attr( 'list', 'reactorvarlist' );
            }
            $inp.attr( 'placeholder', pname );
            if ( undefined !== lastval ) {
                $inp.val( lastval );
            } else {
                $inp.val( parm.default || ( parm.optional ? "" : ( parm.min || 0 ) ) );
            }
        } else {
            if ( "string" !== parm.type ) {
                console.warn("getArgumentField: using default (string) presentation for type " +
                    String(parm.type) + " " + String(pname) );
            }
            $inp = $( '<input class="argument form-control form-control-sm">' );
            if ( ! parm.novars ) {
                $inp.attr( 'list', 'reactorvarlist' );
            }
            $inp.attr( 'placeholder', pname );
            if ( undefined !== lastval ) {
                $inp.val( lastval );
            } else {
                $inp.val( ( undefined === parm.default || parm.optional ) ? "" : parm.default );
            }
        }
        if ( parm.tip ) {
            $inp.attr( 'title', _T( parm.tip ) );  /* localize tip string */
        }

        return $inp;
    }

    function buildActionForm( ad, parameters ) {
        var $ct = $( '<form></form>' );
        var keys = Object.keys( ad.arguments || {} );
        if ( 0 === keys.length ) {
            return false;
        }
        keys.sort( ( a, b ) => {
            let ka = ad.arguments[a].sort || 32767;
            let kb = ad.arguments[b].sort || 32767;
            if ( ka === kb ) {
                if ( a === b ) {
                    return 0;
                }
                return a < b ? -1 : 1;
            }
            return ka < kb ? -1 : 1;
        });
        keys.forEach( pname => {
            var p = ad.arguments[ pname ];
            var id = 'param-' + pname;
            var $af = getArgumentField( "form", pname, p, $ct, ( parameters || {} )[pname] ).attr( 'id', id );
            var $fg = $( '<div class="form-group"></div>' ).appendTo( $ct );
            var label = p.label || pname;
            var extra = [];
            if ( !isEmpty( p.min ) ) {
                extra.push( 'min: ' + String(p.min) );
            }
            if ( !isEmpty( p.max ) ) {
                extra.push( 'max: ' + String(p.max) );
            }
            if ( extra.length ) {
                label += " (" + extra.join( '; ' ) + ")";
            }
            $( "<label></label>" ).attr( "for", id )
                .text( label + ":" )
                .appendTo( $fg );
            $af.appendTo( $fg );
        });
        $( `<div class="action-form-warning">${_T('#action-form-warning')}</div>`).appendTo( $ct );
        return $ct;
    }

    function handleActionClick( event ) {
        var $el = $( event.target );
        var action = $el.data( 'action' ) || "";
        var eid = $el.closest( '.re-listport-item' ).attr( 'id' );
        var entity = api.getEntity( eid );
        let m = action.indexOf( '.' );
        let capName = action.substring( 0, m );
        let cap = entity.getCapability( capName ) || {};
        let ad = ( cap.actions || {} )[ action.substring( m+1 ) ] || {};
        let parm;
        try {
            let last = localStorage.getItem( 'entity_last_action' );
            last = JSON.parse( last );
            let a = last.find( el => el.action == action );
            if ( a ) {
                parm = a.parameters || {};
            }
        } catch ( err ) {
            // nada
        }
        var $form = buildActionForm( ad, parm );
        util.showSysModal( { title: _T("Perform {0}", action ),
            body: $form ? $form :
                _T( 'Ready to perform {0:q} on {1} ({2})?', action, entity.getName(), eid ),
            buttons: [
                { label: _T(['#entity-action-cancel', "Cancel" ]), class: "btn-danger", close: true },
                { label: _T(['#entity-action-perform', "Perform Now" ]), class: "btn-success", ident: "perform" }
            ]
        }).then( function( btn ) {
            if ( "close" !== btn ) {
                var p = {}, ps = {};
                $( '#sysdialog form .argument' ).each( (ix,obj) => {
                    var pname = $(obj).attr( 'id' ).replace( /^param-/, "" );
                    var val = ( $(obj).val() || "" ).trim();
                    ps[ pname ] = val;
                    var pd = ( ad.arguments || {} )[ pname ];
                    var typ = ( pd || {} ).type || "string";
                    if ( "real" === typ ) {
                        val = isNaN( val ) ? null : parseFloat( val );
                    } else if ( typ.match( /^u?int$/ ) || typ.match( /^u?i[1248]$/ ) ) {
                        val = isNaN( val ) ? null : parseInt( val );
                    } else if ( typ.match( /^bool(ean)?/ ) ) {
                        val = null !== val.match( /^(true|yes|1|t|y|on)$/i );
                    }
                    p[ pname ] = val;
                });
                if ( util.hasAnyProperty( p ) ) {
                    let last = localStorage.getItem( 'entity_last_action' );
                    try {
                        last = JSON.parse( last );
                        if ( ! Array.isArray( last ) ) {
                            last = [];
                        }
                    } catch ( err ) {
                        last = [];
                    }
                    let ix = last.findIndex( el => el.action === action );
                    if ( ix < 0 ) {
                        last.push( { action: action, parameters: ps } );
                        if ( last.length > 10 ) {
                            last.splice( 0, 10 - last.length );
                        }
                    } else {
                        let a = last.splice( ix, 1 );
                        a[0].parameters = ps;
                        last.push( a[0] );
                    }
                    localStorage.setItem( 'entity_last_action', JSON.stringify( last ) );
                    //console.log( "entity_last_action", last );
                }
                console.log( "Sending perform", action, p );
                entity.perform( action, p );
            }
            $( '#sysdialog form' ).remove();
        });
        return false;
    }

    function generatePopperContent( e, attr ) {
        const $ul = $( '<ul></ul>' );
        let meta = e.getAttributeMeta( attr );
        if ( meta ) {
            for ( let key in meta ) {
                let vv = meta[ key ];
                let vd = JSON.stringify( vv );
                if ( "number" === typeof vv && vv >= 1528689600000 && vv <= 4133912400000 ) {
                    vd += ` (${new Date(vv).toLocaleString()})`;
                } else if ( vd.length > 64 ) {
                    vd = vd.substring( 0, 61 ) + "...";
                }
                $( '<li></li>' ).text( key + ': ' + vd )
                    .appendTo( $ul );
            }
        } else {
            $ul.append( '<li>No metadata available</li>' );
        }
        return $ul;
    }

    function updateDetailCard( $card, e ) {
        console.assert( $card );
        var eid = e.getCanonicalID();
        var $list = $( 'ul', $card );
        if ( 0 === $list.length ) {
            $list = $( '<ul class="px-0 py-0 re-entity-attrs"></ul>' )
                .attr( 'id', 'attrs-' + eid )
                .appendTo( $card );
        }
        var alist = e.getAttributes();
        var primary = e.getPrimaryAttribute();
        Object.keys( alist || {} ).sort().forEach( function( k ) {
            const spid = ( eid + ":" + k ).replace( /[^a-z0-9]/ig, '_' );
            let $sp = $( 'span#' + idSelector( spid ), $list );
            if ( 0 === $sp.length ) {
                // Should put into alphabetical order???
                let $el = $( '<li></li>' )
                    .appendTo( $list );
                $sp = $( '<span></span>' )
                    .attr( 'id', spid )
                    .text( k + "=" + clean( alist[k].value ) )
                    .toggleClass( 'entity-primary-attribute', k === primary )
                    .appendTo( $el );

                    /* Popper for attribute metadata */
                    try {
                        const $ul = generatePopperContent( e, k );
                        const popper = bootstrap.Popover.getOrCreateInstance( $sp,
                            {
                                title: `${k} metadata`,
                                content: $ul,
                                html: true,
                                delay: 500,
                                placement: 'right',
                                trigger: 'hover focus',
                                container: '#tab-entities',
                                customClass: 're-attr-meta'
                            }
                        );
                        $sp.data( 'popper', popper );
                    } catch ( err ) {
                        console.error( "Failed to create popover for", eid, k, "metadata", err );
                    }
            } else {
                /* Element alread exists */
                $sp.text( k + "=" + clean( alist[k].value ) );
                const popper = $sp.data( 'popper' );
                if ( popper ) {
                    popper.setContent({
                        '.popover-body': generatePopperContent( e, k ).html()
                    });
                }
            }
        });

        var caps = [ ...e.getCapabilities() ];
        var $el = $( 'div#' + idSelector( 'caps-' + eid ), $card );
        if ( 0 === $el.length ) {
            $el = $( '<div class="mt-1"></div>')
                .attr( 'id', 'caps-' + eid )
                .appendTo( $card );
        }
        $el.text( _T( 'Capabilities: {0}', caps.sort().join( ', ' ) ) );

        var acts = [ ...e.getActions() ];
        $el = $( 'div#' + idSelector( 'actions-' + eid ), $card );
        if ( 0 === $el.length ) {
            $el = $( '<div class="mt-1"></div>' )
                .attr( 'id', 'actions-' + eid )
                .appendTo( $card );
        }
        $el.empty().append( _T( 'Actions: ' ) );
        acts.sort().forEach( function( act, ix ) {
            if ( ix > 0 ) {
                $el.append( ', ' );
            }
            $( '<a class="re-actionlink"></a>' ).attr( 'href', '#' )
                .data( 'action', act )
                .text( act )
                .attr( 'title', _T( 'Click to perform this action' ) )
                .appendTo( $el );
        });
        $( 'a.re-actionlink', $el ).on( 'click', handleActionClick );
    }

    function handleEntityActionClick( event ) {
        const $el = $( event.currentTarget );
        const eid = $el.closest( '.re-listport-item' ).attr( 'id' );
        const entity = api.getEntity( eid );
        if ( ! entity ) {
            return;
        }
        switch( $el.data( 'action' ) ) {

            case 'rename':
                const nom = entity.getName() || "";
                let $form = $( '<label></label>' )
                    .append( `<span class="me-1">${_T('#entity-rename-input-label')}:</span>` )
                    .append( $( '<input type="text" class="form-control form-control-sm" >' ).val( nom ) )
                    .append( `<div>${_T('#entity-rename-advice')}</div>` );
                util.showSysModal( { title: _T("#entity-rename-title"),
                    body: $( '<form></form>' ).append( $form ),
                    buttons: [
                        { label: _T([ '#entity-rename-button-cancel', 'Cancel' ]), class: "btn-danger", close: true },
                        { label: _T([ '#entity-rename-button-rename', 'Rename' ]), class: "btn-success" }
                    ]
                }).then( function( btn ) {
                    if ( "close" !== btn ) {
                        const newname = $( '#sysdialog form input' ).val();
                        entity.setName( newname );
                    }
                });
                break;

            case 'delete':
                util.showSysModal( { title: _T("#entity-delete-title"),
                    body: _T([ '#entity-delete-body', 'Deleting an entity removes it from Reactor but does not remove it from the hub/source. If the source object still exists, the entity will be recreated next time the source is inventoried.' ]),
                    buttons: [
                        { label: _T([ '#entity-delete-button-cancel', 'Cancel' ]), class: "btn-success", close: true },
                        { label: _T([ '#entity-delete-button-delete', 'Delete' ]), class: "btn-danger" }
                    ]
                }).then( function( btn ) {
                    if ( "close" !== btn ) {
                        entity.delete();
                        $el.closest( 'div.re-listport-item' ).remove();
                    }
                });
                break;

            case 'copy-attrs':
                const $ul = $el.closest( 'div.re-device-actions' ).siblings( 'ul' );
                var range = document.createRange();
                range.selectNode( $ul.get(0) );
                window.getSelection().removeAllRanges();
                window.getSelection().addRange( range );
                document.execCommand( "copy" );
                window.getSelection().removeAllRanges();
                break;

            default:
                /* nada */
        }
    }

    function showEntityStatusDetail( event ) {
        var $el = $( event.currentTarget );
        var $row = $el.closest( '.re-listport-item' );
        var eid = $row.attr( 'id' );
        var $coll = $( 'div.collapse', $row );
        var $card = $( 'div.card-body', $coll );
        var e = api.getEntity( eid );
        if ( 0 === $card.length ) {
            $( '<div class="w-100"></div>' ).appendTo( $row );
            var $col = $( '<div class="col"></div>' ).appendTo( $row );
            $coll = $( '<div class="collapse show"></div>' ).appendTo( $col );
            $card = $( '<div class="card card-body re-entity-detail bg-transparent mx-0 my-0"></div>' )
                .appendTo( $coll );
            $( `<div class="re-device-actions">
  <button class="btn btn-sm btn-outline-primary" data-action="rename">${_T([ '#entity-rename-button-rename', 'Rename' ])}</button>
  <button class="btn btn-sm btn-outline-danger" data-action="delete">${_T([ '#entity-delete-button-delete', 'Delete' ])}</button>
  <button class="btn btn-sm btn-outline-success" data-action="copy-attrs">${_T([ '#entity-copyattr-button-copy', 'Copy Attributes' ])}</button>
</div>`).appendTo( $card );
            $( '.re-device-actions button[data-action="rename"]' ).prop( 'disabled', ! e.canRename() );
            $( '.re-device-actions button[data-action="delete"]' ).prop( 'disabled', ! e.canDelete() );
            updateDetailCard( $card, e );
            api.subscribe( e, function( event ) {
                // console.log( "entities list detail card notification", event.data.getCanonicalID() );
                if ( "entity-update" === event.type ) {
                    var $item = $( 'div#tab-entities div.re-listport-item[id="' + idSelector( event.data.getCanonicalID() ) + '"]' );
                    var $card = $( 'div.card', $item );
                    // console.log( $item, $card );
                    if ( $card.length > 0 ) {
                        let scrollpos = $( 'ul', $card ).scrollTop();
                        // console.log( "updating card for", event.data.getCanonicalID(), "scrolled to", scrollpos );
                        updateDetailCard( $card, event.data );
                        $( 'ul', $card ).scrollTop( scrollpos );
                    }
                }
            });
            $card.on( 'click', function() { return false; } );
            $( 'div.re-device-actions button', $card ).on( 'click', handleEntityActionClick );

            $( '.re-entity-name i.bi', $row )
                .removeClass( 'bi-chevron-right' )
                .addClass( 'bi-chevron-down' );
        } else {
            /* Because of poppers we now use for attribute meta, we remove the card rather than hiding it.
             * A hidden popper updates continuously and saps performance, and since there is a popper per
             * attribute, there can be a lot of them.
             */
            $( '.re-entity-name i.bi', $row )
                .removeClass( 'bi-chevron-down' )
                .addClass( 'bi-chevron-right' );
            $coll.remove();
        }
        return false;
    }

    var drawFilteredEntities = function( $lp ) {

        var ctrl_filter = $( 'select#filter-ctrl', $lp ).val() || "";
        localStorage.setItem( 'entitylist_filter_ctrl', ctrl_filter );
        var ctrl_re = isEmpty( ctrl_filter ) ? false : new RegExp( `^${ctrl_filter}>`, "i" );
        var grp_filter = $( 'select#filter-group', $lp ).val() || "";
        localStorage.setItem( 'entitylist_filter_grp', grp_filter );
        var grp_entity = isEmpty( grp_filter ) ? false : api.getEntity( grp_filter );
        var cap_filter = $( 'select#filter-cap', $lp ).val() || "";
        localStorage.setItem( 'entitylist_filter_cap', cap_filter );
        var name_filter = $( 'input#filter-name', $lp ).val() || "";
        localStorage.setItem( 'entitylist_filter_name', name_filter );
        var name_re = isEmpty( name_filter ) ? false : new RegExp( name_filter, "i" );

        var filtered = api.getEntities().filter( e => {
            if ( ! ( isEmpty( cap_filter ) || e.hasCapability( cap_filter ) ) ) {
                return false;
            }
            if ( ctrl_re && null === e.getCanonicalID().match( ctrl_re ) ) {
                return false;
            }
            if ( name_re && ! ( name_re.test( e.getName() ) || name_re.test( e.getID() ) ) ) {
                return false;
            }
            if ( grp_entity && ! grp_entity.hasMember( e ) ) {
                return false;
            }
            return true;
        });

        filtered.sort( function( a, b ) {
            return a.getName().localeCompare( b.getName(), undefined, { sensitivity: 'base' } );
        });

        // Get attribute to be displayed (menu of attributes shows when capability filter is active)
        const disp_attr = $( 'select#entity-val-col-selector' ).val() || "";

        var $section = $( '.re-listport-body', $lp ).empty();
        filtered.forEach( function( entity ) {
            var eid = entity.getCanonicalID();
            var $row = util.getListportRow().attr( 'id', eid ).appendTo( $section );
            var $el = $( '<div class="col-xs-12 col-md-6 col-lg-5 col-xl-4 re-entity-name"></div>')
                .appendTo( $row );
            $( '<span></span>' ).text( entity.getName() || eid ).appendTo( $el );
            $( '<i class="bi bi-chevron-right ms-1"></i>' ).appendTo( $el );
            if ( entity.getDeadSince() ) {
                $( '<i class="bi bi-question-diamond-fill ms-3"></i>' )
                    .attr( 'title', _T( ['#entity-missing-alert', 'This entity is marked missing from the hub'] ) )
                    .appendTo( $el );
                $( 'span', $el ).addClass( 'entity-name-deleted' );
            }
            $( '<div class="col overflow-hidden user-select-all"></div>' )
                .text( eid ).attr( 'title', eid )
                .appendTo( $row );
            $el = $( '<div class="col overflow-hidden re-entity-state"></div>' )
                .appendTo( $row );
            if ( "" !== disp_attr ) {
                $el.text( clean( entity.getAttribute( disp_attr ) ) );
            } else if ( null !== entity.getPrimaryAttribute() ) {
                let v = entity.getPrimaryValue();
                $el.text( clean( v ) )
                    .attr( 'title', String(entity.getPrimaryAttribute() ) );
            } else {
                $el.text( '-' );
            }
            var st = entity.getLastUpdate();
            $( '<div class="col text-end overflow-hidden re-entity-update"></div>' )
                .text( null === st ? _T( [ '#entities-never-updated', "never" ] ) : util.shortTime( st ) )
                .appendTo( $row );
        });
        $( 'div.re-entity-name', $section ).on( 'click', showEntityStatusDetail );
    };

    var handleEntityListFilterChange = function( event ) {
        var $lp = $( event.currentTarget ).closest( '.re-listport' );
        // console.log("Redrawing entities (filter changed)");
        drawFilteredEntities( $lp );
    };

    var handleEntityListCapabilityFilterChange = function( ev ) {
        let sel = $( 'select#filter-cap' ).val();
        let $col = $( 'div.col#entities-col-head' );
        if ( "" !== sel ) {
            // "Primary Value" column becomes selector
            let $m = $( '<select></select>' ).attr( 'id', 'entity-val-col-selector' )
                .addClass( 'form-select' )
                .addClass( 'form-select-sm' );
            $( '<option></option>' )
                .val( "" )
                .text( _T(['#entities-col-value','Primary Value']) )
                .appendTo( $m );
            let cap = api.getCapability( sel );
            let attrs = Object.keys( cap?.attributes || {} );
            attrs.sort( (a,b) => a.localeCompare( b, undefined, { sensitivity: 'base' } ) );
            for ( let attr of attrs ) {
                const cattr = `${sel}.${attr}`;
                $( '<option></option>' )
                    .val( cattr )
                    .text( cattr )
                    .appendTo( $m );
            }
            $col.empty().append( $m );
            $m.on( 'change', handleEntityListFilterChange );
        } else {
            // "Primary Value" returns to text.
            $col.empty().text( _T(['#entities-col-value','Primary Value']) );
        }

        // Handle the rest/generic stuff
        handleEntityListFilterChange( ev || { currentTarget: $col } );
    };

    var activate = function( event ) {
        var $tab = $( event.target );
        var $lp = util.getListport().appendTo( $tab );

        var $section = $( ".re-listport-head", $lp );
        var $row = $( '<div class="row"></div>' ).appendTo( $section ); /* not listport-item */
        var $col = $( '<div class="col align-middle"></div>').appendTo( $row );
        $( `<span class="re-listport-title">${_T(['#nav_entities','Entities'])}</span>` ).appendTo( $col );
        $col = $( '<div class="col-auto float-end mt-2"></div>' ).appendTo( $row );
        $( `<button class="btn btn-sm btn-outline-primary ms-2"><i class="bi bi-x me-1"></i><span>${_T(['#ep-filter-clear','Clear Filters'])}</button>` )
            .on( 'click.reactor', function( event ) {
                var $lp = $( event.currentTarget ).closest( '.re-listport' );
                $( 'select.entity_filter' ).val( "" );
                $( 'input#filter-name' ).val( "" );
                $( 'select#filter-cap' ).trigger( 'change' );
            })
            .appendTo( $col );

        $row = $( '<div class="d-flex flex-row mb-2"></div>' ).appendTo( $section ); /* not listport-item */
        $col = $( '<div class=""></div>' ).appendTo( $row );
        $( '<input type="text" id="filter-name" class="form-control form-control-sm" autocomplete="off">' )
            .val( localStorage.getItem( 'entitylist_filter_name' ) || "" )
            .attr( 'placeholder', _T( 'Name Contains:' ) )
            .on( 'keyup', handleEntityListFilterChange )
            .on( 'change', handleEntityListFilterChange )
            .appendTo( $col );
        $col = $( '<div class="ps-1"></div>' ).appendTo( $row );
        var $mm = $( '<select id="filter-ctrl" class="form-select form-select-sm entity_filter"></select>' )
            .on( 'change', handleEntityListFilterChange )
            .appendTo( $col );
        $( `<option value="">${_T( '(any controller)' )}</option>` ).appendTo( $mm );
        var t = api.getControllers();
        t.forEach( (n, ix) => {
            t[ix] = api.getController( n );
        });
        t.sort( (a,b) => {
            return (a.name || a.id).localeCompare( b.name || b.id, undefined, { sensitivity: 'base' } );
        });
        t.forEach( n => {
            $( '<option></option>' ).val( n.id ).text( n.name ? ( n.name + " (" + n.id + ")" ) : n.id )
                .appendTo( $mm );
        });
        $mm.val( localStorage.getItem( 'entitylist_filter_ctrl' ) || "" );
        $col = $( '<div class="ps-1"></div>' ).appendTo( $row );
        $mm = $( '<select id="filter-group" class="form-select form-select-sm entity_filter"></select>' )
            .on( 'change', handleEntityListFilterChange )
            .appendTo( $col );
        $( `<option value="">${_T( '(any group)' )}</option>` ).appendTo( $mm );
        let gl = api.getEntities().filter( e => e.getType() === "Group" );
        gl.sort( (a, b) => {
            return a.getName().localeCompare( b.getName(), undefined, { sensitivity: 'base' } );
        });
        gl.forEach( e => {
            let title = e.getName() + " (" + e.getCanonicalID() + ")";
            $( '<option></option>' ).val( e.getCanonicalID() ).text( title )
                .appendTo( $mm );

        });
        $mm.val( localStorage.getItem( 'entitylist_filter_grp' ) || "" );
        $col = $( '<div class="ps-1"></div>' ).appendTo( $row );
        $mm = $( '<select id="filter-cap" class="form-select form-select-sm entity_filter"></select>' )
            .on( 'change', handleEntityListCapabilityFilterChange )
            .appendTo( $col );
        $( "<option>}</option>" ).val( "" ).text( _T( '(any capability)' ) ).appendTo( $mm );
        let cap = {};
        api.getEntities().forEach( e => {
            let c = e.getCapabilities() || [];
            c.forEach( capName => {
                cap[ capName ] = true;
            });
        });
        cap = Object.keys( cap );
        cap.sort();
        cap.forEach( capName => {
            /* x_vera_svc_micasaverde_com_SceneControllerLED1 => x_vera_..._SceneControlledLED1 */
            let abbrev = capName.length <= 24 ? capName : capName.replace( /^(x_[^_]+_).{4,}(_[^_]+)$/i, "$1...$2" );
            $( '<option></option>' ).val( capName ).text( abbrev ).attr( 'title', capName )
                .appendTo( $mm );
        });
        $mm.val( localStorage.getItem( 'entitylist_filter_cap' ) || "" );

        util.getListportRow().appendTo( $section )
            .append( `<div class="col-xs-12 col-md-6 col-lg-5 col-xl-4">${_T(['#entities-col-name','Entity Name'])}</div>` )
            .append( `<div class="col">${_T(['#entities-col-id','Canonical ID'])}</div>` )
            .append( `<div id="entities-col-head" class="col">${_T(['#entities-col-value','Primary Value'])}</div>` )
            .append( `<div class="col text-end">${_T(['#entities-col-modified','Last Modified'])}</div>` );

        $( 'select#filter-cap' ).trigger( 'change' );

        $( window ).resize( function() { util.listportResize( $lp, 42 ); } ).trigger( 'resize' );

        api.off( "." + tabId );
        api.on('entity_change.' + tabId, function( entity ) {
            // console.log("reactor-ui-core: entity change id " + entity.getCanonicalID() + " name " + String(entity.getName()));
            var $row = $( 'div#' + idSelector( entity.getCanonicalID() ) );
            $( '.re-entity-name span', $row ).text( entity.getName() );
            $( '.re-entity-state', $row ).text( clean( entity.getPrimaryValue() ) );
            $( '.re-entity-update', $row ).text( util.shortTime( entity.getLastUpdate(), "%F %T" ) );
            /* Detail card, if present, takes care of itself */
            $row.stop().clearQueue().css( 'background-color', '#0c0' );
            $row.animate( { backgroundColor: '#fff' }, 5000);
        });
        api.on( 'entity_delete.' + tabId, function( entity ) {
            var $row = $( 'div#' + idSelector( entity.getCanonicalID() ) );
            $row.remove();
        });

        $( 'a#testlink' ).attr( 'href', '#' ).on( 'click', function() {
            api.recycle();
        });
    };

    var suspend = function( event ) {
        // console.log("reactor-ui-entities.suspend() running");
        // console.log(event);
        var $tab = $( event.target );
        if ( tabId !== $tab.attr( 'id' ) || !$tab.hasClass( "re-tab-container" ) ) {
            // console.log( "    event is not for the tab, but a child" );
            return;
        }

        $tab.empty();

        api.off( '.' + tabId );

        return true;
    };

    return {
        "init": function( $main ) {
            $( 'div#' + idSelector( tabId ) + '.re-tab-container' ).remove();
            var $tab = $( '<div></div>' )
                .hide()
                .attr( 'id', tabId )
                .addClass( 're-tab-container' )
                .appendTo( $main );
            $tab.on( 'activate', activate );
            $tab.on( 'suspend', suspend );

            /* Our styles. */
            if ( 0 === $('link#reactor-entities-styles').length ) {
                $('head').append('<link id="reactor-entities-styles" href="lib/css/reactor-ui-entities.css" rel="stylesheet">');
            }

        },
        "tab": function() {
            return $( 'div#' + idSelector( tabId ) + ".re-tab-container" );
        }
    };
})(jQuery);
