/** This file is part of Reactor.
 *  Copyright (c) 2020-2024 Kedron Holdings LLC, All Rights Reserved.
 *  Reactor is not public domain or open source. Distribution or derivative works are expressly prohibited.
 */

export const version = 24119;

import api from '/client/ClientAPI.js';

import ReactorUICore from "./reactor-ui-core.js";

import { getLocale } from './i18n.js';

function readyHandler() {
    console.log("reactor-ui.js:readyHandler()");

    let cookies = document.cookie.split( /;\s*/ ) || [];
    let cookie = cookies.find( el => el.match( /^reactor-session=/ ) );  // set by reaction-login.js
    let authorization;
    if ( cookie ) {
        let t = cookie.split( /=/ )[1];
        let parts = t.split( /\|/ );
        authorization = parts[1];
        if ( parts[2] ) {
            const applications = parts[2].split(',');
            if ( 1 === applications.length && "true" === applications[0] ) {
                // All apps OK
            } else if ( ! applications.includes( 'reactor' ) ) {
                window.location.href = "login.html?msg=Refused";
                return;
            }
        }
    }

    api.off( 'auth_fail.core' ).on( 'auth_fail.core', () => {
        window.location.href = "login.html";
    });
    api.start( { authorization: authorization } ).then( () => {
        ReactorUICore.start();
    }).catch ( err => {
        if ( err.status === 401 ) {
            window.location.href = "login.html";
        } else {
            console.error( err );
        }
    });

    let $el = jQuery( 'a#re-sysclock' );
    let update_browser_time = function() {
        /* Fixed format time display */
        let t = ( new Date() ).toLocaleTimeString( getLocale(), { hour12: false } );
        if ( t !== $el.text() ) {
            $el.text( t );
        }
        setTimeout( update_browser_time, 1000 - ( Date.now() % 1000 ) );
    };
    // Kick it off.
    setTimeout( update_browser_time, 500 );
}

jQuery( readyHandler );
